from typing import Dict, Iterable, TypeVar, Union, Any

K = TypeVar("K")
V = TypeVar("V")

def omit(obj: Union[Dict[K, V], Any], keys: Iterable[K]) -> Dict[K, V]:
    """Returns a new dict omitting the specified keys.
    
    Supports both dictionaries and objects with a __dict__ attribute.
    If the object does not have a __dict__, raises a TypeError.
    """
    keys_set = set(keys)

    if isinstance(obj, dict):
        return {k: v for k, v in obj.items() if k not in keys_set}

    if hasattr(obj, "__dict__"):
        return {k: v for k, v in vars(obj).items() if k not in keys_set}

    raise TypeError("omit() expects a dict or an object with a __dict__ attribute.")



from typing import Callable, Literal, TypedDict
from auth0_ai.authorizers.types import ToolInput

class ToolCallContext(TypedDict):
    thread_id: str
    tool_call_id: str
    tool_name: str

AuthContext = Literal["tool-call", "tool", "thread", "agent"]

ContextGetter = Callable[ToolInput, ToolCallContext]

def ns_from_context(auth_context: AuthContext, call_context: ToolCallContext) -> list[str]:
    """
    Resolve a namespace based on the credential-sharing scope (auth_context).

    Args:
        auth_context (AuthContext): The context in which credentials are scoped.
        call_context (ToolCallContext): Information about the current tool call.

    Returns:
        list[str]: A list of namespace components.
    """
    thread_ns = ["threads", call_context["thread_id"]]
    tool_ns = ["tools", call_context["tool_name"]]
    tool_call_ns = ["tool_calls", call_context["tool_call_id"]]

    match auth_context:
        case "tool-call":
            return thread_ns + tool_ns + tool_call_ns
        case "tool":
            return thread_ns + tool_ns
        case "thread":
            return thread_ns
        case "agent":
            return []



from abc import ABC
from typing import Any, Type, TypeVar, get_type_hints
from auth0_ai.interrupts.auth0_interrupt import Auth0Interrupt
from auth0_ai.authorizers.async_authorization import AsyncAuthorizationRequest

class WithRequestData:
    def __init__(self, request: AsyncAuthorizationRequest):
        self._request = request

    @property
    def request(self) -> AsyncAuthorizationRequest:
        return self._request

AsyncAuthorizationInterruptType = TypeVar("T", bound="AsyncAuthorizationInterrupt")

class AsyncAuthorizationInterrupt(Auth0Interrupt, ABC):
    def __init__(self, message: str, code: str):
        super().__init__(message, code)

    @classmethod
    def is_interrupt(cls: Type[AsyncAuthorizationInterruptType], interrupt: Any) -> bool:
        return (
            interrupt
            and Auth0Interrupt.is_interrupt(interrupt)
            and isinstance(interrupt["code"], str)
            and (
                (hasattr(cls, "code") and getattr(cls, "code") == interrupt["code"])
                or
                (not hasattr(cls, "code") and interrupt["code"].startswith("ASYNC_AUTHORIZATION_"))
            )
        )

    @classmethod
    def has_request_data(cls, interrupt: Any) -> bool:
        if not cls.is_interrupt(interrupt):
            return False

        if not isinstance(interrupt, dict):
            return False

        request = interrupt.get("_request")
        if not isinstance(request, dict):
            return False

        required_keys = set(get_type_hints(AsyncAuthorizationRequest).keys())
        return required_keys <= request.keys()


class AccessDeniedInterrupt(AsyncAuthorizationInterrupt, WithRequestData):
    code: str = "ASYNC_AUTHORIZATION_ACCESS_DENIED"

    def __init__(self, message: str, request: AsyncAuthorizationRequest):
        AsyncAuthorizationInterrupt.__init__(self, message, AccessDeniedInterrupt.code)
        WithRequestData.__init__(self, request)


class UserDoesNotHavePushNotificationsInterrupt(AsyncAuthorizationInterrupt):
    code: str = "ASYNC_AUTHORIZATION_USER_DOES_NOT_HAVE_PUSH_NOTIFICATIONS"

    def __init__(self, message: str):
        super().__init__(message, UserDoesNotHavePushNotificationsInterrupt.code)


class AuthorizationRequestExpiredInterrupt(AsyncAuthorizationInterrupt, WithRequestData):
    code: str = "ASYNC_AUTHORIZATION_REQUEST_EXPIRED"

    def __init__(self, message: str, request: AsyncAuthorizationRequest):
        AsyncAuthorizationInterrupt.__init__(self, message, AuthorizationRequestExpiredInterrupt.code)
        WithRequestData.__init__(self, request)


class AuthorizationPendingInterrupt(AsyncAuthorizationInterrupt, WithRequestData):
    code: str = "ASYNC_AUTHORIZATION_PENDING"

    def __init__(self, message: str, request: AsyncAuthorizationRequest):
        AsyncAuthorizationInterrupt.__init__(self, message, AuthorizationPendingInterrupt.code)
        WithRequestData.__init__(self, request)

    def next_retry_interval(self) -> int:
        """Return the interval in seconds to wait before the next retry attempt."""
        return self.request["interval"]


class AuthorizationPollingInterrupt(AsyncAuthorizationInterrupt, WithRequestData):
    code: str = "ASYNC_AUTHORIZATION_POLLING_ERROR"

    def __init__(self, message: str, request: AsyncAuthorizationRequest, retry_after: int = None):
        Auth0Interrupt.__init__(self, message, AuthorizationPollingInterrupt.code)
        WithRequestData.__init__(self, request)
        self.retry_after = retry_after

    def next_retry_interval(self) -> int:
        """Return the interval in seconds to wait before the next retry attempt."""
        # Use the retry_after value from the HTTP header if available,
        # otherwise fall back to the original interval from the auth request
        return self.retry_after if self.retry_after is not None else self.request["interval"]


class InvalidGrantInterrupt(AsyncAuthorizationInterrupt, WithRequestData):
    code: str = "ASYNC_AUTHORIZATION_INVALID_GRANT"

    def __init__(self, message: str, request: AsyncAuthorizationRequest):
        Auth0Interrupt.__init__(self, message, InvalidGrantInterrupt.code)
        WithRequestData.__init__(self, request)



from typing import Any, Callable, Generic, Optional, Sequence, TypeVar, TypedDict, Union
from auth0_ai.stores.store import Store, StorePutOptions

T = TypeVar("T")
TNew = TypeVar("TNew")

class SubStoreParams(TypedDict, total=False):
    base_namespace: Sequence[str]
    get_ttl: Callable[[T], int | None]

class SubStore(Generic[T], Store[T]):
    """
    A store that wraps a parent store, optionally prefixing namespaces and deriving TTL values from a user-defined function.
    """

    def __init__(self, parent: Store[Any], options: Optional[SubStoreParams[T]] = None):
        if parent is None:
            raise ValueError("Parent store is required")

        self._parent = parent
        self._base_namespace = options["base_namespace"] if options and "base_namespace" in options else []
        self._get_ttl = options["get_ttl"] if options and "get_ttl" in options else None

    def _full_namespace(self, namespace: Sequence[str]) -> list[str]:
        return list(self._base_namespace) + list(namespace)

    async def get(self, namespace: Sequence[str], key: str) -> T | None:
        return await self._parent.get(self._full_namespace(namespace), key)

    async def delete(self, namespace: Sequence[str], key: str) -> None:
        await self._parent.delete(self._full_namespace(namespace), key)

    async def put(
        self,
        namespace: Sequence[str],
        key: str,
        value: T,
        options: Optional[StorePutOptions] = None
    ) -> None:
        expires_in = (
            options["expires_in"] if options and "expires_in" in options else
            self._get_ttl(value) if self._get_ttl else None
        )

        put_options = {"expires_in": expires_in} if expires_in is not None else None

        await self._parent.put(
            self._full_namespace(namespace),
            key,
            value,
            put_options
        )

    def create_sub_store(
        self,
        options: Union[
            str,
            Sequence[str],
            SubStoreParams[TNew]
        ] = None
    ) -> "SubStore[TNew]":
        base_namespace: Optional[Sequence[str]] = None
        get_ttl: Optional[Callable[[TNew], int | None]] = None

        if isinstance(options, str):
            base_namespace = [options]
        elif isinstance(options, list):
            base_namespace = options
        elif isinstance(options, dict):
            base_namespace = options.get("base_namespace")
            get_ttl = options.get("get_ttl")

        return SubStore[TNew](self, {
            "base_namespace": base_namespace or [],
            "get_ttl": get_ttl
        })



import asyncio
from typing import Generic, TypedDict, Optional, Union, Tuple, Callable, TypeVar, Awaitable, ParamSpec

class Auth0ClientParams(TypedDict):
    """
    Base authorizer parameters.

    Attributes:
        domain (str): The domain of your Auth0 tenant.
        client_id (str): Your application's client ID.
        client_secret (str, optional): Your application's client secret.
        client_assertion_signing_key (str, optional): Private key used to sign the client assertion JWT.
        client_assertion_signing_alg (str, optional): Algorithm used to sign the client assertion JWT (defaults to 'RS256').
        telemetry (bool, optional): Enable or disable telemetry (defaults to True).
        timeout (float or tuple, optional): Change the requests connect and read timeout. Pass a tuple to specify both values separately or a float to set both to it (defaults to 5.0 for both).
        protocol (str, optional): Useful for testing (defaults to 'https').
    """
    domain: Optional[str]
    client_id: Optional[str]
    client_secret: Optional[str]
    client_assertion_signing_key: Optional[str]
    client_assertion_signing_alg: Optional[str]
    telemetry: Optional[bool]
    timeout: Optional[Union[float, Tuple[float, float]]]
    protocol: Optional[str]

ToolInput = ParamSpec("ToolInput")
TResult = TypeVar("TResult")

class AuthorizerToolParameter(Generic[ToolInput, TResult]):
    def __init__(self, value: Union[
        TResult,
        Callable[ToolInput, Union[TResult, Awaitable[TResult]]]
    ]):
        self.value = value

    async def resolve(
        self,
        *args: ToolInput.args,
        **kwargs: ToolInput.kwargs,
    ) -> TResult:
        result = self.value(*args, **kwargs) if callable(self.value) else self.value
        return await result if asyncio.iscoroutine(result) else result



from typing import TypedDict

class AsyncAuthorizationRequest(TypedDict):
    """
    Attributes:
        id (str): The authorization request ID. Use this ID to check the status of the authorization request.
        requested_at (float): The timestamp (in seconds since the Unix epoch) when the authorization request was made.
        expires_in (int): The time in seconds for the authorization request to expire.
        interval (int): The interval to use when polling the status.
    """
    id: str
    requested_at: float
    expires_in: int
    interval: int



import asyncio
import contextvars
import hashlib
import inspect
import json
import os
import time
from contextlib import asynccontextmanager
from datetime import datetime
from typing import Any, Callable, Dict, Generic, Optional, Sequence, TypedDict, Union
from auth0 import Auth0Error
from auth0.authentication.back_channel_login import BackChannelLogin
from auth0.authentication.get_token import GetToken
from auth0_ai.credentials import TokenResponse
from auth0_ai.authorizers.async_authorization.async_authorizer_params import AsyncAuthorizerParams
from auth0_ai.authorizers.async_authorization.async_authorization_request import AsyncAuthorizationRequest
from auth0_ai.authorizers.types import Auth0ClientParams, ToolInput
from auth0_ai.interrupts.async_authorization_interrupts import AccessDeniedInterrupt, AuthorizationPendingInterrupt, AuthorizationPollingInterrupt, AuthorizationRequestExpiredInterrupt, InvalidGrantInterrupt, UserDoesNotHavePushNotificationsInterrupt
from auth0_ai.stores import SubStore, InMemoryStore
from auth0_ai.authorizers.context import ns_from_context, ContextGetter
from auth0_ai.utils import omit

class AsyncStorageValue(TypedDict):
    context: Any
    credentials: Optional[TokenResponse]
    # The namespace in the Store for the Async Auth authorization response.
    auth_request_ns: Sequence[str];

_local_storage: contextvars.ContextVar[Optional[AsyncStorageValue]] = contextvars.ContextVar("local_storage", default=None)

def _get_local_storage() -> AsyncStorageValue:
    store = _local_storage.get()
    if store is None:
        raise RuntimeError("The tool must be wrapped with the with_async_authorization function.")
    return store

def _update_local_storage(data: AsyncStorageValue) -> None:
    store = _get_local_storage()
    updated = store.copy()
    updated.update(data)
    _local_storage.set(updated)

@asynccontextmanager
async def _run_with_local_storage(data: AsyncStorageValue):
    if _local_storage.get() is not None:
        raise RuntimeError("Cannot nest tool calls that require Async Authorization.")
    token = _local_storage.set(data)
    try:
        yield
    finally:
        _local_storage.reset(token)

def get_async_authorization_credentials() -> TokenResponse | None:
    local_store = _get_local_storage()
    return local_store.get("credentials")

def _ensure_openid_scope(scopes: list[str]) -> str:
    if "openid" not in scopes:
        scopes.insert(0, "openid")
    return " ".join(scopes)

class AsyncAuthorizerBase(Generic[ToolInput]):
    def __init__(self, params: AsyncAuthorizerParams[ToolInput], auth0: Auth0ClientParams = None):
        auth0 = {
            "domain": (auth0 or {}).get("domain", os.getenv("AUTH0_DOMAIN")),
            "client_id": (auth0 or {}).get("client_id", os.getenv("AUTH0_CLIENT_ID")),
            "client_secret": (auth0 or {}).get("client_secret", os.getenv("AUTH0_CLIENT_SECRET")),
            "client_assertion_signing_key": (auth0 or {}).get("client_assertion_signing_key"),
            "client_assertion_signing_alg": (auth0 or {}).get("client_assertion_signing_alg"),
            "telemetry": (auth0 or {}).get("telemetry"),
            "timeout": (auth0 or {}).get("timeout"),
            "protocol": (auth0 or {}).get("protocol")
        }

        # Remove keys with None values
        auth0 = {k: v for k, v in auth0.items() if v is not None}

        self.back_channel_login = BackChannelLogin(**auth0)
        self.get_token = GetToken(**auth0)
        self.auth0 = auth0
        self.params = params

        # TODO: consider moving this to Auth0AI classes
        async_authorization_store = SubStore(params["store"] if "store" in params else InMemoryStore()).create_sub_store("AUTH0_AI_ASYNC_AUTHORIZATION")

        self.auth_request_store = SubStore[AsyncAuthorizationRequest](async_authorization_store, {
            "get_ttl": lambda auth_request: auth_request["expires_in"] * 1000 if "expires_in" in auth_request else None
        })

        self.credentials_store = SubStore[TokenResponse](async_authorization_store, {
            "get_ttl": lambda credential: credential["expires_in"] * 1000 if "expires_in" in credential else None
        })

    def _handle_authorization_interrupts(self, err: Union[AuthorizationPendingInterrupt, AuthorizationPollingInterrupt]) -> None:
        raise err

    def _get_instance_id(self, authorize_params) -> str:
        props = {
            "auth0": omit(self.auth0, ["client_secret", "client_assertion_signing_key"]),
            "params": authorize_params
        }
        sh = json.dumps(props, sort_keys=True, separators=(",", ":"))
        return hashlib.md5(sh.encode("utf-8")).hexdigest()

    async def _get_authorize_params(self, *args: ToolInput.args, **kwargs: ToolInput.kwargs) -> Dict[str, Any]:
        authorize_params = {
            "scope": _ensure_openid_scope(self.params.get("scopes")),
            "audience": self.params.get("audience"),
            "requested_expiry": self.params.get("requested_expiry"),
        }

        if isinstance(self.params.get("user_id"), str):
            user_id = self.params.get("user_id")
        elif inspect.iscoroutinefunction(self.params.get("user_id")):
            user_id = await self.params.get("user_id")(*args, **kwargs)
        else:
            user_id = self.params.get("user_id")(*args, **kwargs)

        if not user_id:
            raise ValueError("Unable to resolve user id, check the provided user_id parameter.")

        authorize_params["login_hint"] = f'{{ "format": "iss_sub", "iss": "https://{self.back_channel_login.domain}/", "sub": "{user_id}" }}'

        if isinstance(self.params.get("binding_message"), str):
            authorize_params["binding_message"] = self.params.get("binding_message")
        elif inspect.iscoroutinefunction(self.params.get("binding_message")):
            authorize_params["binding_message"] = await self.params.get("binding_message")(*args, **kwargs)
        else:
            authorize_params["binding_message"] = self.params.get("binding_message")(*args, **kwargs)

        if isinstance(self.params.get("authorization_details"), list):
            authorize_params["authorization_details"] = json.dumps(self.params.get("authorization_details"))
        elif inspect.iscoroutinefunction(self.params.get("authorization_details")):
            authorize_params["authorization_details"] = json.dumps(await self.params.get("authorization_details")(*args, **kwargs))
        elif callable(self.params.get("authorization_details")):
            authorize_params["authorization_details"] = json.dumps(self.params.get("authorization_details")(*args, **kwargs))

        return authorize_params

    async def _start(self, authorize_params) -> AsyncAuthorizationRequest:
        requested_at = time.time()

        try:
            response = self.back_channel_login.back_channel_login(**authorize_params)
            return AsyncAuthorizationRequest(
                id=response["auth_req_id"],
                requested_at=requested_at,
                expires_in=response["expires_in"],
                interval=response["interval"],
            )
        except Auth0Error as e:
            if e.error_code == "invalid_request":
                raise UserDoesNotHavePushNotificationsInterrupt(e.message)
            else:
                raise

    def _extract_retry_after_header(self, error: Auth0Error) -> Optional[int]:
        """
        Extract the Retry-After header value from an Auth0Error.
        
        Args:
            error: The Auth0Error object that may contain HTTP headers
            
        Returns:
            The retry-after value in seconds as an integer, or None if not present
        """

        if not hasattr(error, 'headers') or not error.headers:
            return None
            
        retry_after = error.headers.get('retry-after') or error.headers.get('Retry-After')
        
        if retry_after is None:
            return None
            
        try:
            return int(retry_after)
        except (ValueError, TypeError):
            # If the retry-after value is not a valid integer, return None
            return None

    def _get_credentials_internal(self, auth_request: AsyncAuthorizationRequest) -> TokenResponse | None:
        try:
            # Calculate elapsed time in seconds
            elapsed_seconds = datetime.now().timestamp() - auth_request["requested_at"]

            if elapsed_seconds >= auth_request["expires_in"]:
                raise AuthorizationRequestExpiredInterrupt(
                    "The authorization request has expired.",
                    auth_request
                )

            response = self.get_token.backchannel_login(auth_req_id=auth_request["id"])
            return TokenResponse(
                access_token=response["access_token"],
                expires_in=response["expires_in"],
                scope=response.get("scope", "").split(),
                token_type=response.get("token_type"),
                id_token=response.get("id_token"),
                refresh_token=response.get("refresh_token"),
            )

        except Auth0Error as e:
            if e.error_code == "authorization_pending":
                raise AuthorizationPendingInterrupt(e.message, auth_request)

            if e.error_code == "slow_down":
                retry_after = self._extract_retry_after_header(e)
                raise AuthorizationPollingInterrupt(e.message, auth_request, retry_after)

            if e.error_code == "invalid_grant":
                raise InvalidGrantInterrupt(e.message, auth_request)

            if e.error_code == "invalid_request":
                raise UserDoesNotHavePushNotificationsInterrupt(e.message)

            if e.error_code == "access_denied":
                raise AccessDeniedInterrupt(e.message, auth_request)

            raise

    def _get_credentials(self, auth_request: AsyncAuthorizationRequest) -> TokenResponse | None:
        return self._get_credentials_internal(auth_request)

    async def get_credentials_polling(self, auth_request: AsyncAuthorizationRequest) -> TokenResponse | None:
        credentials: TokenResponse | None = None

        while not credentials:
            try:
                credentials = self._get_credentials_internal(auth_request)
            except (AuthorizationPendingInterrupt, AuthorizationPollingInterrupt) as err:
                await asyncio.sleep(err.next_retry_interval())
            except Exception:
                raise

        return credentials

    async def delete_auth_request(self):
        local_store = _get_local_storage()
        auth_request_ns = local_store["auth_request_ns"]
        await self.auth_request_store.delete(auth_request_ns, "auth_request")

    def protect(
        self,
        get_context: ContextGetter[ToolInput],
        execute: Callable[ToolInput, any]
    ) -> Callable[ToolInput, any]:
        async def wrapped_execute(*args: ToolInput.args, **kwargs: ToolInput.kwargs):
            context = get_context(*args, **kwargs)
            authorize_params = await self._get_authorize_params(*args, **kwargs)
            instance_id = self._get_instance_id(authorize_params)
            auth_request_ns = [instance_id, "auth_requests", *ns_from_context("tool-call", context)]
            credentials_ns = [instance_id, "credentials", *ns_from_context(self.params.get("credentials_context", "tool-call"), context)]

            local_store = {
                "context": context,
                "auth_request_ns": auth_request_ns,
            }

            async with _run_with_local_storage(local_store):
                interrupt_mode = self.params.get("on_authorization_request", "interrupt") == "interrupt"

                try:
                    credentials = await self.credentials_store.get(credentials_ns, "credential")
                    if not credentials:
                        if interrupt_mode:
                            auth_request = await self.auth_request_store.get(auth_request_ns, "auth_request")
                            if not auth_request:
                                # initial request
                                auth_request = await self._start(authorize_params)
                                await self.auth_request_store.put(auth_request_ns, "auth_request", auth_request)

                            credentials = self._get_credentials(auth_request)
                        else:
                            # block mode
                            auth_request = await self._start(authorize_params)
                            credentials = await self.get_credentials_polling(auth_request)

                        await self.delete_auth_request()

                        if credentials is not None:
                            await self.credentials_store.put(credentials_ns, "credential", credentials)
                except (AuthorizationPendingInterrupt, AuthorizationPollingInterrupt) as interrupt:
                    return self._handle_authorization_interrupts(interrupt)
                except Exception as err:
                    await self.delete_auth_request()
                    raise

                _update_local_storage({"credentials": credentials})

                if inspect.iscoroutinefunction(execute):
                    return await execute(*args, **kwargs)
                else:
                    return execute(*args, **kwargs)

        return wrapped_execute



import asyncio
import time
from typing import Dict, TypeVar, Sequence
from auth0_ai.stores.store import Store, StorePutOptions

T = TypeVar("T")

class InMemoryStore(Store[T]):
    """
    An in-memory store for dev/demo purposes.
    """
    
    def __init__(self):
        self._store: Dict[str, tuple[T, float | None]] = {}
        self._lock = asyncio.Lock()

    def _get_key(self, namespace: Sequence[str], key: str) -> str:
        return "/".join(namespace) + "/" + key

    async def get(self, namespace: Sequence[str], key: str) -> T | None:
        store_key = self._get_key(namespace, key)

        async with self._lock:
            item = self._store.get(store_key)
            if item is None:
                return None

            value, expires_at = item
            now = time.time() * 1000  # milliseconds

            if expires_at is not None and now >= expires_at:
                del self._store[store_key]
                return None

            return value

    async def delete(self, namespace: Sequence[str], key: str) -> None:
        store_key = self._get_key(namespace, key)

        async with self._lock:
            self._store.pop(store_key, None)

    async def put(
        self,
        namespace: Sequence[str],
        key: str,
        value: T,
        options: StorePutOptions | None = None
    ) -> None:
        store_key = self._get_key(namespace, key)
        expires_in = options["expires_in"] if options and options.get("expires_in") is not None else None
        expires_at = time.time() * 1000 + expires_in if expires_in is not None else None

        async with self._lock:
            self._store[store_key] = (value, expires_at)



from typing import Final, Type, TypeVar, Dict, Any

Auth0InterruptType = TypeVar("T", bound="Auth0Interrupt")

class Auth0Interrupt(Exception):
    _name: Final[str] = "AUTH0_AI_INTERRUPT"

    def __init__(self, message: str, code: str):
        super().__init__(message)
        self.code = code

    @property
    def name(self) -> str:
        return self._name

    def to_json(self) -> Dict[str, Any]:
        """
        Serialize the interrupt to a JSON object.
        """
        return {
            key: value for key, value in self.__dict__.items()
        } | {"message": self.args[0], "name": self.name}

    @classmethod
    def is_interrupt(cls: Type[Auth0InterruptType], interrupt: Any) -> bool:
        """
        Checks if an interrupt is of a specific type asserting its data component.
        """
        if isinstance(interrupt, dict):
            return (
                interrupt.get("name") == "AUTH0_AI_INTERRUPT"
                and (not hasattr(cls, "code") or interrupt.get("code") == getattr(cls, "code", None))
            )
        
        return (
            isinstance(interrupt, Auth0Interrupt)
            and interrupt.name == "AUTH0_AI_INTERRUPT"
            and (not hasattr(cls, "code") or interrupt.code == getattr(cls, "code", None))
        )
